/* AMX Mod X
*   Hostage Voices
*
* (c) Copyright 2006-2007 by VEN
*
* This file is provided as is (no warranties)
*
*     DESCRIPTION
*       Plugin allow to set a different custom voices for each hostage.
*
*     NOTE
*       The plugin "Hostage Models" can be used as a supplement.
*
*     FEATURES
*       - configuration file support
*       - per-map configs support
*       - voice on headshot kill
*       - disable hostages' voice
*
*     CONFIGURATION
*       There are a $configsdir/hostage_voices/default.ini file that is
*       used to load hostages' custom voices for all maps that contain
*       hostages. It's also possible to load per-map configs. E.g. if
*       map with a name "mymap" is loaded, firstly plugin will try to
*       load mymap.ini config from the "hostage_voices" directory and
*       if it isn't found the default.ini file will be loaded instead.
*       There are a compilation options that is allow to change path to
*       the configuration files and default config file name/extension.
*       Also there are a compilation options that is related to the
*       internal data of the configuration files. Compilation options
*       can be altered in the plugin's source code. To read more about
*       configuring and for examples see default configuration file.
*
*     CVARS
*       hv_mode (N: mode, default: 1) - controls plugin's mode
*         1 - enable custom hostages voices
*         2 - also custom voice on headshot kill
*         0 - use game default hostages voices
*        -1 - disable all hostages voices
*
*     VERSIONS
*       0.2   added configuration file support/per-map configs support
*             added CVar to control implemented modes of the plugin
*             fixed possible issue for non-default phrases number value
*       0.1.2 disabled custom voice on headshot kill (can be enabled)
*             code reorganized, minor efficiency/technique improvements
*       0.1.1 fixed error on hostage headshot kill
*       0.1   first release
*
*     CREDITS
*       pl2003 - plugin request/initial idea
*/

// plugin's main information
#define PLUGIN_NAME "Hostage Voices"
#define PLUGIN_VERSION "0.2"
#define PLUGIN_AUTHOR "VEN"

#include <amxmodx>
#include <fakemeta>

// OPTIONS BELOW

// you can set "hardcoded" or amxx_*-based path (see core.ini)
#define CONFIG_FILE_PATH "|configsdir|/hostage_voices"
// by default "|configsdir|" would be "addons/amxmodx/configs"

// will be used if separate config for the current map isn't found
#define CONFIG_FILE_NAME "default"
#define CONFIG_FILE_EXTN "ini"

// you can increase these values if necessary
#define MAX_CUSTOM_HOSTAGES 4
#define MAX_HOSTAGE_PHRASES 5
#define MAX_PHRASE_FILENAME_LEN 95

// OPTIONS ABOVE

new g_hostages_phrases[MAX_CUSTOM_HOSTAGES][MAX_HOSTAGE_PHRASES][MAX_PHRASE_FILENAME_LEN + 1]

new g_loaded_hostages_num = 0
new g_loaded_phrases_num[MAX_CUSTOM_HOSTAGES] = {0, ...}

new const g_headshot_sound_prefix[] = "player/headshot"

new g_hostages_ent[MAX_CUSTOM_HOSTAGES]
new g_hostages_num = 0

new g_pcvar_mode

// load and precache hostages custom phrases
public plugin_precache() {
	load_config()

	new loaded_phrases_num, phrase_first_char
	for (new i = 0; i < g_loaded_hostages_num; ++i) {
		loaded_phrases_num = g_loaded_phrases_num[i]
		for (new j = 0; j < loaded_phrases_num; ++j) {
			phrase_first_char = g_hostages_phrases[i][j][0]
			if (phrase_first_char != '^0' && phrase_first_char != '*')
				precache_sound(g_hostages_phrases[i][j])
		}
	}
}

public plugin_init() {
	register_plugin(PLUGIN_NAME, PLUGIN_VERSION, PLUGIN_AUTHOR)

	g_pcvar_mode = register_cvar("hv_mode", "1")

	new ent = FM_NULLENT
	new const classname[] = "classname"
	new const hostage_entity[] = "hostage_entity"
	while (g_hostages_num != g_loaded_hostages_num && (ent = engfunc(EngFunc_FindEntityByString, ent, classname, hostage_entity)))
		// to improve performance cache hostage entities indeces
		g_hostages_ent[g_hostages_num++] = ent

	// if we don't have a hostages on this map
	if (!engfunc(EngFunc_FindEntityByString, FM_NULLENT, classname, hostage_entity)) {
		pause("d") // stop this plugin
		return
	}

	// while speaking hostages is emit a sounds, hook it
	register_forward(FM_EmitSound, "forward_emit_sound")
}

public forward_emit_sound(ent, channel, const sample[], Float:volume, Float:attenuation, flags, pitch) {
	static mode
	mode = get_pcvar_num(g_pcvar_mode)
	// if plugin is disabled or entity isn't valid do not go further
	if (!mode || !pev_valid(ent))
		return FMRES_IGNORED

	static bool:is_headshot
	is_headshot = false
	// if it's a headshot sound
	if (equali(sample, g_headshot_sound_prefix, sizeof g_headshot_sound_prefix - 1)) {
		is_headshot = true

		// if custom voice on headshot feature is disbaled do not go further
		if (mode < 2)
			return FMRES_IGNORED
	}

	// if "disable hostage voices" mode is enabled
	if (mode < 0) {
		if (is_headshot)
			return FMRES_IGNORED

		static classname[32]
		pev(ent, pev_classname, classname, sizeof classname - 1)
		static hostage_entity[] = "hostage_entity"
		// if it's a hostage
		if (equal(classname, hostage_entity))
			// disable hostage voice
			return FMRES_SUPERCEDE
	}

	static i
	// if it isn't one of the custom hostages do not go further
	if ((i = get_hostageent_arrayelem_index(ent)) == -1)
		return FMRES_IGNORED

	static loaded_phrases_num
	// if no loaded phrases for this hostage do not go further
	if ((loaded_phrases_num = g_loaded_phrases_num[i]) == 0)
		return FMRES_IGNORED

	static j
	// retrieve a random phrase index
	j = random(loaded_phrases_num)

	static phrase_start_char
	phrase_start_char = g_hostages_phrases[i][j][0]
	// if empty phrase
	if (phrase_start_char == '^0')
		// disable hostage voice
		return FMRES_SUPERCEDE
	else if (phrase_start_char == '*')
		// allow original hostage voice
		return FMRES_IGNORED

	// emit our custom voice
	engfunc(EngFunc_EmitSound, ent, channel, g_hostages_phrases[i][j], volume, attenuation, flags, pitch)

	// supercede original voice
	return FMRES_SUPERCEDE
}

get_hostageent_arrayelem_index(ent) {
	static i
	for (i = 0; i < g_hostages_num; ++i) {
		if (g_hostages_ent[i] == ent)
			return i
	}

	return -1
}

load_config() {
	new file_path[128] = CONFIG_FILE_PATH
	new file_path_real[sizeof file_path] = CONFIG_FILE_PATH
	new buffer[2048] = "amxx_"

	new const left = contain(file_path, "|")
	if (left != -1) {
		new const right = contain(file_path[left + 1], "|") + left + 1
		if (right != left) {
			file_path[left] = '^0'
			file_path[right] = '^0'

			new const len = strlen(buffer)
			formatex(buffer[len], strlen(file_path[left + 1]), "%s", file_path[left + 1])
			get_localinfo(buffer, buffer, sizeof file_path - 1)

			formatex(file_path_real, sizeof file_path_real - 1, "%s%s%s", file_path[0], buffer, file_path[right + 1])
		}
	}

	if (!dir_exists(file_path_real))
		return

	new mapname[32]
	get_mapname(mapname, sizeof mapname - 1)
	formatex(buffer, sizeof file_path_real - 1, "%s/%s.%s", file_path_real, mapname, CONFIG_FILE_EXTN)
	if (!file_exists(buffer))
		formatex(buffer, sizeof file_path_real - 1, "%s/%s.%s", file_path_real, CONFIG_FILE_NAME, CONFIG_FILE_EXTN)

	new const file = fopen(buffer, "rt")
	if (!file)
		return

	new buffer0, phrase[MAX_PHRASE_FILENAME_LEN + 1], last_char_pos, phrase_file[MAX_PHRASE_FILENAME_LEN + 1]
	while (g_loaded_hostages_num != MAX_CUSTOM_HOSTAGES && fgets(file, buffer, sizeof buffer - 1)) {
		buffer0 = buffer[0]
		if (buffer0 == '^n' || buffer0 == ';' || buffer0 == '#' || (buffer0 == '/' && buffer[1] == '/'))
			continue

		new phrases = 0
		while (phrases != MAX_HOSTAGE_PHRASES && buffer[0]) {
			strbreak(buffer, phrase, MAX_PHRASE_FILENAME_LEN, buffer, sizeof buffer - 1)
			last_char_pos = strlen(phrase) - 1
			if (phrase[last_char_pos] == '?') {
				phrase[last_char_pos] = '^0'
				formatex(phrase_file, MAX_PHRASE_FILENAME_LEN, "sound/%s", phrase)
				if (!file_exists(phrase_file))
					continue
			}

			g_hostages_phrases[g_loaded_hostages_num][phrases++] = phrase
		}

		g_loaded_phrases_num[g_loaded_hostages_num++] = phrases
	}

	fclose(file)
}
